#if !defined(__UNICODE_FUNC_HEADER__)
#define __UNICODE_FUNC_HEADER__

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Unicode Related Function
//   version: 0.0.0.2
//   description: this header was written to deal with unicode (wide char)
//   author: hobbyscripter <hobbyscripter@gmail.com>
//   copyright: licensed under NSIS-license
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


// ===============
// Data Type
// ===============
typedef struct {
  int charCount;
  unsigned char byte[3];
} utf8_char;


// ===============
// Function Declare
// ===============
void initialUTF8Char(utf8_char* target);
void convertUnicode16ToUTF8(utf8_char* target, wchar_t source);
void appendUTF8CharToStr(char* target, utf8_char* source);


// ===============
// Function Realization
// ===============
// ---------------
// Initial `utf8_char'
// ---------------
void
initialUTF8Char(utf8_char* target){
  target->charCount = 0;

  target->byte[0] = '\0';
  target->byte[1] = '\0';
  target->byte[2] = '\0';
}

// ---------------
// Convert Char from Unicode-16 to UTF-8
// ---------------
void
convertUnicode16ToUTF8(utf8_char* target,
                       wchar_t source)
{
  // ~~~~~ Reformat `source' ~~~~~
  unsigned char highByte;
  unsigned char lowByte;

  highByte = source >> 8;
  lowByte = source & 0xFF;

  // ~~~~~ Init `target' ~~~~~
  initialUTF8Char(target);
  
  // ~~~~~ Process ~~~~~
  // `0x0000 ~ 0x007F':
  if (highByte == 0 && lowByte < 0x8F){
    target->charCount = 1;
    target->byte[0] = lowByte;
    return;
  }

  // `0x0080 ~ 0x07FF':
  if ((highByte == 0 && lowByte >= 0x80) ||
      (highByte > 0 && highByte < 0x08))
  {
    target->charCount = 2;
    target->byte[0] = (0xC0 | (highByte << 2) | (lowByte >> 6));
    target->byte[1] = (0x80 | (lowByte & 0x3F));
    return;
  }

  // `0x0800 ~ 0xFFFF':
  if (highByte >= 0x08){
    target->charCount = 3;
    target->byte[0] = (0xE0 | highByte >> 4);
    target->byte[1] = (0x80 | ((highByte & 0xF) << 2) | (lowByte >> 6));
    target->byte[2] = (0x80 | (lowByte & 0x3F));
    return;
  }
}

void
appendUTF8CharToStr(char* target,
                    utf8_char* source)
{
  // ~~~~~ WARNING!! NOTICE!! ~~~~~
  // DO make sure the `target' is big enough to store a new utf8_char.
  // DONT pass a char-array as a `target'

  // ~~~~~ Go to the End of the String ~~~~~
  int pos; pos = 0;
  while (target[pos] != '\0'){pos++;} // Get Position of the First '\0'
  
  // ~~~~~ Copy UTF-8 Char ~~~~~
  int i;
  for (i = 0; i < (source->charCount); i++){
    target[pos + i] = (char)source->byte[i];
    
  }
  target[pos + source->charCount] = '\0'; // Add an NULL for End
}


#endif // !defined(__UNICODE_FUNC_HEADER__)
// EOF

