// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Firefox Portable String Convert Plugin
//   version 0.0.0.2
//   description: this file is the source of the plugin
//   author: hobbyscripter <hobbyscripter@gmail.com>
//   copyright: licensed under NSIS-license
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


// ===============
// Include
// ===============
#include <windows.h>
#include <stdlib.h>

#include "../include/multibyte.h"
#include "../include/utils.h"
#include "../include/unicode.h"
#include "../include/exdll.h"


// ===============
// Global Variables
// ===============
HINSTANCE g_hInstance;
HWND g_hwndParent;


// ===============
// Function
// ===============
// ---------------
// Convert String To UTF-8
// ---------------
void
__declspec(dllexport)
ConvertStringToUTF8(HWND hwndParent,
                    int string_size,
                    char *variables,
                    stack_t **stacktop,
                    extra_parameters *extra)
{
  g_hwndParent=hwndParent;

  EXDLL_INIT();
  {
    // ----- Create Buffer -----
    char buff[1024]; INIT_ARRAY(buff, 1024, '\0');

    // ----- Get NSISMaxStrLen -----
    int NSISMaxStrLen;
    int NSISMaxStrLenSafe;

    popstring(buff);
    NSISMaxStrLen = atoi(buff);

    // `Make-Sure-NSISMaxStrLen-Is-Bigger-Than-Zero':
    NSISMaxStrLen = (NSISMaxStrLen <= 0)?1024:NSISMaxStrLen;

    NSISMaxStrLenSafe = NSISMaxStrLen - 1; // Preserve One for '\0'

    // ----- Alloc -----
    char* inputString; inputString = NULL;
    wchar_t* unicodeString; unicodeString = NULL; // For Unicode 16
    char* outputString; outputString = NULL;

    inputString = (char*)malloc(sizeof(char) * NSISMaxStrLen);
    unicodeString = (wchar_t*)malloc(sizeof(wchar_t) * NSISMaxStrLen);
    outputString = (char*)malloc(sizeof(char) * NSISMaxStrLen);

    INIT_ARRAY(inputString, NSISMaxStrLen, '\0');
    INIT_ARRAY(unicodeString, NSISMaxStrLen, 0x0000);
    INIT_ARRAY(outputString, NSISMaxStrLen, '\0');

    // ----- Get Input String -----
    popstring(inputString);

    // ----- Convert the String to Wide Char (via Win32API) -----
    int convToU16Return;
    convToU16Return = MultiByteToWideChar(CP_ACP, 0,
                                          inputString, NSISMaxStrLen,
                                          unicodeString, NSISMaxStrLen);

    // `Check if return errors':
    if (convToU16Return == 0){
      pushstring(inputString);
      goto ConvertStringToUTF8_FuncUninit; // JUMP to `FunctionUninitLabel'
    }
    
    // ----- Convert to UTF-8 and copy to outputString -----
    // Note: Using this method to convert to UTF-8 instead of using Windows
    //       API, because Windows 95/98/ME DOESNT support UTF-8 Codepage
    int pos; pos = 0;
    int wpos;
    utf8_char tempChar;
    for (wpos = 0; wpos < NSISMaxStrLen; wpos++){
      if (unicodeString[wpos] == 0x0000){
        // `End-Of-Source-String':
        break; // Just break is ok, because '\0' is filled in when init
      }
      else{
        convertUnicode16ToUTF8(&tempChar, unicodeString[wpos]);

        if (pos + tempChar.charCount >= NSISMaxStrLenSafe){
          // `End-Of-Copy-Buffer':
          break; // Just break is ok, because '\0' is filled in when init
        }
        else{
          appendUTF8CharToStr(outputString, &tempChar);
          pos += tempChar.charCount;
        }
      }
    }

    // ----- Push Output -----
    pushstring(outputString);

ConvertStringToUTF8_FuncUninit:
    // ----- Free Memory -----
    free(inputString);
    free(unicodeString);
    free(outputString);

    // ----- Return -----
    return;
  }
}

// ---------------
// Escape String With '%' (like what does Firefox do)
// ---------------
void
__declspec(dllexport)
EscapeString(HWND hwndParent,
             int string_size,
             char *variables,
             stack_t **stacktop,
             extra_parameters *extra)
{
  g_hwndParent=hwndParent;

  EXDLL_INIT();
  {
    // ----- Create Buffer -----
    char buff[1024]; INIT_ARRAY(buff, 1024, '\0');

    // ----- Get NSISMaxStrLen -----
    int NSISMaxStrLen;
    int NSISMaxStrLenSafe;

    popstring(buff);
    NSISMaxStrLen = atoi(buff);

    // `Make-Sure-NSISMaxStrLen-Is-Bigger-Than-Zero':
    NSISMaxStrLen = (NSISMaxStrLen <= 0)?1024:NSISMaxStrLen;

    NSISMaxStrLenSafe = NSISMaxStrLen - 1; // Preserve One for '\0'

    // ----- Alloc -----
    int ipos; ipos = 0; // For the `for' loop (input)
    int opos; opos = 0; // For the `for' loop (output)
    char* inputString; inputString = NULL;
    char* outputString; outputString = NULL;

    inputString = (char*)malloc(sizeof(char) * NSISMaxStrLen);
    outputString = (char*)malloc(sizeof(char) * NSISMaxStrLen);

    INIT_ARRAY(inputString, NSISMaxStrLen, '\0');
    INIT_ARRAY(outputString, NSISMaxStrLen, '\0');

    // ----- Create Table -----
    const int escapeTargetChar[256] = {
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0,
      1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1,
      
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
    };
    const char* hexChars = "0123456789ABCDEF";

    // ----- Get Input String -----
    popstring(inputString);

    // ----- Change '\' Directory Split into '/' -----
    int previousCharIsLowByte; previousCharIsLowByte = 0;
    // `Handle-inputString[0]':
    if (inputString[0] == '\\'){
      inputString[0] = '/';
    }
    // `Handle-Other':
    for (ipos = 1; ipos < NSISMaxStrLen; ipos++){
      if (previousCharIsLowByte){
        // To Prevent the Code like this (%AD%B1)\ Convert to (%AD%B1)%5C
        // This should be convert if it is '\\', because there won't
        // be two low byte at once
        if (inputString[ipos] == '\\'){
          inputString[ipos] = '/';
        }

        previousCharIsLowByte = 0; // Clear the label
      }
      else if (isHighByteOfMultiByte(inputString[ipos - 1])){
        // This Character is the Second Byte of `MultiBytes' Char
        // Give it a label, and DONT convert even if it's '\\'
        previousCharIsLowByte = 1;
      }
      else{ // else if (!isHighByteOfMultiByte(inputString[ipos - 1])){
        // This Character is Not the Second Byte of `MultiBytes' Char
        
        if (inputString[ipos] == '\\'){
          inputString[ipos] = '/';
        }

        previousCharIsLowByte = 0; // Reset the label
      }
    }

    // ----- DEVELOPER-NOTE -----
    // For the escape convert, we are flowing the rules below:
    //   1. Escape every char whose escapeTargetChar[c] returns true.
    //   2. Escape '|' when previous word is multi byte head

    // ----- Convert Now -----
    unsigned char currentChar;
    unsigned char prevChar; prevChar = 0;
    for (ipos = 0; ipos < NSISMaxStrLen; ipos++){
      currentChar = (unsigned char)inputString[ipos];

      if (currentChar == '\0'){
        // `End-Of-Input-String':
        break;
      }
      
      if (escapeTargetChar[currentChar] ||
          (currentChar == '|' && prevChar > 0x7F))
      {
        // Need to escape

        if (opos + 3 <= NSISMaxStrLenSafe){
          outputString[opos] = '%';
          outputString[opos + 1] = hexChars[currentChar >> 4];
          outputString[opos + 2] = hexChars[currentChar & 0xF];
          opos += 3;
        }
        else {
          // `End-Of-Output-Buffer':
          break;
        }
      }
      else{
        if (opos + 1 <= NSISMaxStrLenSafe){
          outputString[opos] = (char)currentChar;
          opos++;
        }
        else{
          // `End-Of-Output-Buffer':
          break;
        }
      }

      prevChar = currentChar;
    }

    // ----- Push Output -----
    pushstring(outputString);

    // ----- Free Memory -----
    free(inputString);
    free(outputString);

    // ----- Return -----
    return;
  }
}


BOOL WINAPI DllMain(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
  g_hInstance=hInst;
  return TRUE;
}
